/*
 * XLife Copyright 1989 Jon Bennett jb7m+@andrew.cmu.edu, jcrb@cs.cmu.edu
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * CMU SUCKS
 */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <stdlib.h>
#include "defs.h"

#define	char	unsigned char
#include "icon.h"
#include "cursor_data.h"
#include "cursor_mask.h"
#undef char

extern char *getenv();

#ifndef GRAB_FRACTION
#define GRAB_FRACTION	0.8	/* fraction of screen to seize */
#endif /* GRAB_FRACTION */

static coord_t savex, savey;	/* last point seen during boxing */

static char *fixedcolors[] = {
    "Black",
    "White",
    "Brown",
    "Blue",
    "Green",
    "#cb0000" /* lighted red */,
    "Yellow",
    "Purple",
    "Orange",
    "Magenta",
    "Cyan",
    "Grey",
    "Pink",
    "Plum",
    "#ADFF2F", /* lightgreen */
    "Tan",
    "#ecb088", /* faded red */
    "#444444",
    "#777777",
    "Red"
};

static long int xmarker[3], ymarker[3];
static XColor cellcolor[MAXCOLORS];
static XGCValues xgcv;

void DoExpose(const int win) {
    if (convmode) return;
    switch(win) {
    case LIFEWIN:
	while (XCheckMaskEvent(disp, ExposureMask, &event));
	XClearWindow(disp, lifew);
        showgrid = 1;
	redrawscreen();
	break;
    case INPUTWIN:
	XClearWindow(disp, inputw);
	XDrawString(disp, inputw, ntextgc, ICOORDS(0,0), inpbuf, 
                    strlen(inpbuf));
	break;
    case RULEWIN:
	showrules();
	break;
    case HELPWIN:
        if (helpw_mode == -1)
	   redraw_help();
        else if (helpw_mode < 0)
           redraw_viewvars();
        else
           eo_comments = redraw_comments();
    }
}

int colorstates(void) {
   if (maxstates == 2) return 0;
   return maxstates;
}

void ResizeLW(int oheight) {
   inputwwidth = width - RULEW - dispcoord*(COORDW + BORDERWIDTH)
                    - BORDERWIDTH*2 - colorstates()*(FONTWIDTH + 3);
   if (inputwwidth < 30) inputwwidth = 30;
   XResizeWindow(disp, inputw, inputwwidth, INPUTH);
   XResizeWindow(disp, rulew, RULEW + colorstates()*(FONTWIDTH + 3),
                  oheight - BORDERWIDTH);
   XMoveWindow(disp, rulew,
                  width - RULEW - COORDW - BORDERWIDTH*2 - colorstates()*(FONTWIDTH + 3),
                  oheight - BORDERWIDTH);
}

void DoResize(void) {
    int owidth = width, oheight = height;

    width = event.xconfigure.width;
    height = event.xconfigure.height;
    xpos += shr((owidth - width)/2, scale);
    ypos += shr((oheight - height)/2, scale);
    inputlength = width/FONTWIDTH;
    oheight = height - INPUTH - BORDERWIDTH*3;
    if (oheight < 5) oheight = 5;
    ResizeLW(oheight);
    XResizeWindow(disp, lifew, width - BORDERWIDTH*2, oheight);
    XResizeWindow(disp, helpw, width - BORDERWIDTH*2, height - BORDERWIDTH*2);
    if (helpw_mode || eo_comments) DoExpose(HELPWIN);
    XMoveWindow(disp, inputw, 0, oheight - BORDERWIDTH);
    XMoveWindow(disp, coordw, width - COORDW - BORDERWIDTH*2,
                            oheight - BORDERWIDTH);
    showgrid = 1;
    redrawscreen();
}

void alloc_states(unsigned int newmaxstates) {
    if (newmaxstates == maxstates) return;
    confirmload();
/*    if (newmaxstates < maxstates)
          clear_all();*/
    clipboard_copy(&active);
    if (newmaxstates == 2)
          changesize(sizeof(struct twostate_t));
    else if (ev_mode == GEN_DRIVEN)
          changesize(sizeof(struct gstate_t));
    else
          changesize(sizeof(struct nstate_t));
    maxstates = newmaxstates;
    clipboard_flush(&active, maxstates);
    ResizeLW(height - INPUTH - BORDERWIDTH*3);
}

int DoKeySymIn(KeySym keysym) {
    switch (keysym) {
    case XK_4:
    case XK_KP_4:
    case XK_Left:
	xpos -= shr((unsigned)width/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_6:
    case XK_KP_6:
    case XK_Right:
	xpos += shr((unsigned)width/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_8:
    case XK_KP_8:
    case XK_Up:
	ypos -= shr((unsigned)height/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_2:
    case XK_KP_2:
    case XK_Down:
	ypos += shr((unsigned)height/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_7:
    case XK_KP_7:
	xpos -= shr((unsigned)width/4, scale);
	ypos -= shr((unsigned)height/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_9:
    case XK_KP_9:
	xpos += shr((unsigned)width/4, scale);
	ypos -= shr((unsigned)height/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_3:
    case XK_KP_3:
	xpos += shr((unsigned)width/4, scale);
	ypos += shr((unsigned)height/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_1:
    case XK_KP_1:
	xpos -= shr((unsigned)width/4, scale);
	ypos += shr((unsigned)height/4, scale);
	DoExpose(LIFEWIN);
	break;
    case XK_5:
    case XK_KP_5:
	center();
	break;
    case XK_0:
    case XK_KP_0:
	median();
	break;
    case XK_Help:
	help();
	break;
    default:
	return 0;
    }

    /* could process it */
    return 1;
}

void inc_scale(void) {
   if (scale < 6) {
            setscale(++scale);
            xpos += XPOS(event.xmotion.x, 0);
            ypos += YPOS(event.xmotion.y, 0);
            DoExpose(LIFEWIN);
   }
}

void dec_scale(void) {
   if (scale > -10) {
            xpos -= XPOS(event.xmotion.x, 0);
            ypos -= YPOS(event.xmotion.y, 0);
            setscale(--scale);
            DoExpose(LIFEWIN);
   }
}

void DoKeyIn(char *kbuf) {
    int num;

    switch(kbuf[0]) {
    case 'r':
        showgrid = 1;
	redrawscreen();
	break;

    case 'R':
	newrules();
        displaystats();
	break;

    case 'F':
	loadrules();
	break;

    case '=':
    case '+':
        inc_scale();
	break;

    case '-':
        dec_scale();
	break;

    case '.':
	xpos += shr(event.xbutton.x - width/2, scale);
	ypos += shr(event.xbutton.y - height/2, scale);
	DoExpose(LIFEWIN);
	break;

    case 'g':
	confirmload();
	state = state == RUN ? STOP : RUN;
	DoExpose(LIFEWIN);
	break;

    case '@':
        if (maxstates == 2 || active_rules[strlen(active_rules) - 1] == '+') {
           char s[100];

           confirmload();
           strcpy(s, active_rules);
           if (s[strlen(s) - 1] == '+')
              s[strlen(s) - 1] = 0;
           else
              strcat(s, "+");
           file_rules(s);
           DoExpose(LIFEWIN);
        }
        else {
           strcpy(inpbuf, "History is only allowed for 2-state automata");
           DoExpose(INPUTWIN);
        }
        break;

    case 'c':
        bounding_box(&active);
	dispboxes ^= 1;
        displaystats();
	break;

    case '$':
        bounding_box(&active);
        dispchanges ^= 1;
        displaystats();
        break;

    case '*':
        multicolor ^= 1;
        pseudocolor = 0;
        DoExpose(LIFEWIN);
        break;

    case '~':
        changesonly ^= 1;
        pseudocolor = 0;
        DoExpose(LIFEWIN);
        break;

    case 'E':
        if ((pseudocolor = (pseudocolor + 1)%3) > 0)
           if (pseudocolor == 1)
              pseudopaint[0][1] = 0;
           else
              pseudopaint[0][1] = 2;
        multicolor = changesonly = 0;
        DoExpose(LIFEWIN);
        break;

    case 'o':
        num = tentative.tiles != 0;
	confirmload();
	generate(&active);
        if (num)
           DoExpose(LIFEWIN);
        else
	   redisplay();
	break;

    case 'p':
	if (dispcoord)
	    dispcoord = FALSE;
	else
	{
	    lastx--;	/* fake move to force coordinates to print */
	    dispcoord = TRUE;
	}
	/* force resize of input window */
	XResizeWindow(disp, inputw,
		      width - RULEW - dispcoord*(COORDW + BORDERWIDTH)
                        - BORDERWIDTH*2, INPUTH);
	break;

    case 'n':
	current_tile();
	break;

    case '(':
    case '[':
    case '{':
	for (num = 0; kbuf[0] != "([{"[num]; num++);
	xmarker[num] = XPOS(event.xmotion.x, xpos);
	ymarker[num] = YPOS(event.xmotion.y, ypos);
	sprintf(inpbuf, "Marker %d set to (%ld,%ld)", 
		num, xmarker[num] - xorigin, ymarker[num] - yorigin);
	DoExpose(INPUTWIN);
	break;

    case ')':
    case ']':
    case '}':
	for (num = 0; kbuf[0] != ")]}"[num]; num++);
	xpos += xmarker[num] - XPOS(event.xmotion.x, xpos);
	ypos += ymarker[num] - XPOS(event.xmotion.y, ypos);
	DoExpose(LIFEWIN);
    break;

    case 'J': {
	long x, y;

        strcpy(inpbuf, "Jump to x, y: ");
        minbuflen = 14;
        DoExpose(INPUTWIN);
        getxstring();
	if (sscanf(inpbuf + 13, "%ld,%ld", &x, &y) == 2) {
	    xpos += x + xorigin - XPOS(event.xmotion.x, xpos);
	    ypos += y + yorigin - YPOS(event.xmotion.y, ypos);
	    DoExpose(LIFEWIN);
	}
      }
      break;

    case 'T':
        confirmload();
        strcpy(inpbuf, "Enter topology: ");
        minbuflen = 16;
        DoExpose(INPUTWIN);
        getxstring();
        if (!inpbuf[16]) {
             displaystats();
             break;
        }
        num = limits;
        if (set_topology(inpbuf + 16) > 0) {
             strncpy(inpbuf, " Wrong topology ", 16);
             DoExpose(INPUTWIN);
             announce_and_wait();
             displaystats();
             break;
        }
        if (bounding_box(&active) && limits && (active.xmin < x_min_limit && x_min_limit
              || active.xmax > x_max_limit && x_max_limit
              || active.ymin < y_min_limit && y_min_limit
              || active.ymax > y_max_limit && y_max_limit || num != limits) ) {
             make_tentative(active.xmin, active.ymin, active.xmax, active.ymax);
             clear_pattern(&active);
             center();
        }
        else {
           tile *cptr = active.tiles, *nptr; /* keep correct screen */

           for (; cptr; cptr = nptr) {
             nptr = cptr->next;
             if (cptr->x < x_min_limit || cptr->x > x_max_limit
                || cptr->y < y_min_limit || cptr->y > y_max_limit)
                killtile(&active, cptr);
           }
        }
        DoExpose(LIFEWIN);
        break;

    case 'i':
	saveformat = 'M';
	strcpy(inpbuf, "Image format for saving patterns: A D I R M P S (default = M): ");
        minbuflen = 63;
        DoExpose(INPUTWIN);
        getxstring();
	if (inpbuf[63] && strchr("ADIRMPS", inpbuf[63]))
	     saveformat = inpbuf[63];
        /*XMaskEvent(disp, KeyPressMask, &event);
        XLookupString(&event.xkey, keybuf, 16, &ks, &kstat);
        if (strchr("ADIRMPS", keybuf[0]))
             saveformat = keybuf[0];*/
	break;

    case 'j':
        switch (hideperiod) {
        case 0:
           hideperiod = 50;
           break;
        case 50:
           hideperiod = 200;
           break;
        default:
           hideperiod = 0;
        }
        if (hideperiod < 2)
           strcpy(inpbuf, "No jump");
        else
           sprintf(inpbuf, "Jump set to %u", hideperiod);
        DoExpose(INPUTWIN);
        break;

    case 'O':
	announce("Origin set to active cell");
	xorigin = XPOS(event.xmotion.x,xpos);
	yorigin = YPOS(event.xmotion.y,ypos);
	break;

    case 'C':
        if (tentative.tiles)
           clear();
        else {
           setscale(scale = 3);
           clear();
           center();
           cellmass = 0;
        }
	break;

    case 'S':
	savefile();
        showgrid = 1;
	redrawscreen();
	break;

    case 'W':
	saveloadscript();
	free_loadscript();
	break;

    case 'D':
	free_loadscript();
	if (tentative.tiles) {
	    undoload();
	    DoExpose(LIFEWIN);
	    announce("Load script (and latest load) discarded");
	}
	else
	    announce("Load script discarded"); 
	break;

    case 'l':
        stashed[0] = 0;
	loadfile();
        sprintf(inpbuf, "%d lines of comments", numcomments);
        DoExpose(INPUTWIN);
        adjhash(&tentative);
        break;

    case 'L':
        if (stashed[0])
           loadfile();
        else
           announce("No previous load");
        break;

    case 'u':
        txx = tyy = 1;
        txy = tyx = 0;
        loadx = iloadx;
        loady = iloady;
        DoExpose(LIFEWIN);
        break;

    case 'h':
	confirmload();
	if (state == HIDE) {
	    state = STOP;
	    DoExpose(LIFEWIN);
	}
	else
	    state = HIDE;
	break;

    case '/':
        if (bounding_box(&active) == 0)
            sprintf(inpbuf, "Life is extinct");
        else if (maxstates != 2) {
            inpbuf[0] = 0;
            for (num = 1; num < maxstates; num++)
               sprintf(inpbuf + strlen(inpbuf), " %d", active.cellcount[num]);
        }
        announce_and_wait();
        displaystats();
        break;

    case 'b':
        if (bounding_box(&active) == 0)
            sprintf(inpbuf, "Life is extinct");
        else
            sprintf(inpbuf,
                    "Life bounds: %ld <= x <= %ld  %ld <= y <= %ld",
                    active.xmin - xorigin, active.xmax - xorigin,
                    active.ymin - yorigin, active.ymax - yorigin);
        announce_and_wait();
        displaystats();
        break;

    case '?':
	help();
	break;

    case 'f':
	settimeout(delay = DELAY_FAST);
	break;

    case 'm':
	settimeout(delay = DELAY_MED);
	break;

    case 's':
	settimeout(delay = DELAY_SLOW);
	break;

    case '>':
        if (delay >= DELAY_INCREMENT)
            settimeout(delay -= DELAY_INCREMENT);
        break;

    case '<':
        settimeout(delay += DELAY_INCREMENT);
        break;

    case '!':
	randomize();
	break;

    case '%': {
           float r;

           strcpy(inpbuf, "Input density (% or nothing): ");
           minbuflen = 31;
           DoExpose(INPUTWIN);
           getxstring();
           if (sscanf(inpbuf + 30, "%f", &r) == 1)
               rnd_density = r/100;
           sprintf(inpbuf, "Using %.2f%% density", rnd_density*100);
           DoExpose(INPUTWIN);
        }
        break;

    case '#': /* toggle wireframe tentative pattern mode */
        wireframe ^= 1;
        if (tentative.tiles)
             DoExpose(LIFEWIN);
        break;

    case 'M': /* toggle gridmode */
        dispmesh ^= 1;
        DoExpose(LIFEWIN);
        break;

    case '^': {
	   unsigned x = 0;

           strcpy(inpbuf, "Set random seed (integer number): ");
           minbuflen = 17;
           DoExpose(INPUTWIN);
           getxstring();
	   if (sscanf(inpbuf + 16, "%u", &x) == 1)
	       srandom(randomseed = x);
	   sprintf(inpbuf, "Random seed is set to %u", x);
	   DoExpose(INPUTWIN);
        }
        break;

    case 'N':
	name_file();
	break;

    case 'A':
	comment();
	break;

    case 'V':
	view_comments();
	break;

    case 'K':
        numcomments = outcome[0] = 0;
        announce("Comments are discarded");
        break;

    case 'v':
        viewvars();
        break;

    case 'B':
	benchmark();
	break;

    case 'Q':
	exit(0);

    case 'U': /* Get rid of loaded pattern */
	undoload();
	DoExpose(LIFEWIN);
	break;

    case 'I': /* Force confirm of loaded pattern */
	confirmload();
	DoExpose(LIFEWIN);
	break;

    case 'd':
	copy_tentative();
        DoExpose(LIFEWIN);
	break;

    case 'G': /* perform some generations on loaded pattern */
        genload();
	DoExpose(LIFEWIN);
	break;

    case 'x':
       if (tentative.tiles) {
         long dlx, dly;

         clipboard_copy(&tentative);
         loadx += dlx = XPOS(event.xbutton.x, xpos) - loadx;
         loady += dly = YPOS(event.xbutton.y, ypos) - loady;
         clipboard.x -= (dly*txy - dlx*tyy)/(txy*tyx-txx*tyy);
         clipboard.y -= (dlx*tyx - dly*txx)/(txy*tyx-txx*tyy);
         clear_pattern(&tentative);
         clipboard_flush(&tentative, maxstates);
         DoExpose(LIFEWIN);
       }
       break;

    case 'a':
        if (ev_mode == TABLE_DRIVEN)
	  set_transition();
	break;

    case 't':
        if (ev_mode == TABLE_DRIVEN)
	  test_transition();
	break;

    default:
	break;
    }
    kbuf[0] = '\0';	/* get rid of old keystroke so shift doesn't bring it back */
}

void Motion(void) { /* handle X motion events */
    if (event.xmotion.window == lifew) {
        coord_t x = XPOS(event.xmotion.x, xpos), y = YPOS(event.xmotion.y, ypos);
        int c;

        if (event.xmotion.state & Button1MotionMask && (!limits || chk_limits(x, y))) {
            c = lookcell(&active, x, y);
            drawcell(event.xmotion.x, event.xmotion.y, paintcolor);
            chgcell(&active, x, y, paintcolor);
            if (c != paintcolor) {
                        active.cellcount[paintcolor]++;
                        active.cellcount[c]--;
                        displaystats();
            }
        }
        else if (event.xmotion.state & Button3MotionMask) {
            /* erase the old box, draw the new one */
            erasebox(loadx, loady, savex, savey);
            if (!limits)
              drawbox(loadx, loady, savex = x, savey = y, 1);
            else if (chk_limits(loadx, loady)) {
                if (limits&1)
                  if (x < x_min_limit)
                    x = x_min_limit;
                  else if (x >= x_max_limit)
                    x = x_max_limit - 1;
                if (limits&2)
                  if (y < y_min_limit)
                    y = y_min_limit;
                  else if (y >= y_max_limit)
                    y = y_max_limit - 1;
                drawbox(loadx, loady, savex = x, savey = y, 1);
            }
        }
        else if (event.xmotion.state & Button2MotionMask && (!limits || chk_limits(x, y))) {
                c = lookcell(&active, x, y);
                drawcell(event.xmotion.x, event.xmotion.y, 0);
                chgcell(&active, x, y, 0);
                active.cellcount[c]--;
                displaystats();
            }
    }
}

void Button(void) { /* handle a button-press event */
    if (event.xbutton.window == rulew)
	setcolor(-1, event.xbutton.x, event.xbutton.y);
    else if (event.xbutton.window == lifew) {
        switch(event.xbutton.button) {
          case 4:
            inc_scale();
            return;
          case 5:
            dec_scale();
            return;
          }
        if (tentative.tiles)
          switch(event.xbutton.button) {
          case 1:
            moveload();
            break;
          case 2:
            flipload();
            break;
          case 3:
            turnload();
          }
        else {
          int c;
          coord_t x = XPOS(event.xmotion.x, xpos), y = YPOS(event.xmotion.y, ypos);

	  switch(event.xbutton.button) {
	  case 1:
            if (!limits || chk_limits(x, y)) {
              c = lookcell(&active, x, y);
              drawcell(event.xbutton.x, event.xbutton.y, paintcolor);
	      chgcell(&active, x, y, paintcolor);
	      showcoord(TRUE);
	      if (c != paintcolor) {
			active.cellcount[paintcolor]++;
			active.cellcount[c]--;
                        displaystats();
	      }
            }
	    break;
	  case 2:
            if (!limits || chk_limits(x, y)) {
              c = lookcell(&active, x, y);
	      drawcell(event.xbutton.x, event.xbutton.y, 0);
	      chgcell(&active, x, y, 0);
	      showcoord(TRUE);
	      active.cellcount[c]--;
              displaystats();
            }
            break;
	  case 3:
            savex = loadx = x;
	    savey = loady = y;
	  }
        }
    }
}

static void Release() {
    if (event.xbutton.window == lifew && !tentative.tiles
				&& event.xbutton.button == 3) {
       if (loadx == savex && savey == loady) {
	   xpos += shr(event.xbutton.x - width/2, scale);
	   ypos += shr(event.xbutton.y - height/2, scale);
	}
	else {
	   make_tentative(loadx, loady, savex, savey);
           changesonly = 0;
        }
        DoExpose(LIFEWIN);
    }
}

main(int argc, char **argv) {
    int i, tmp, qgens = 0;
    Cursor cursor;
    Pixmap icon, cursor_data, cursor_mask;
    XSizeHints hints;
    XWMHints wm_hints;
    XClassHint class_hints;
    XSetWindowAttributes winat;
    XColor white, black;
    char *geomstring = NULL;
    char *initpat = NULL;
    struct timeval inputdelay, timeouttemp;
    LoadReq *loadqueue = NULL;
    XTextProperty windowName, iconName;
    unsigned long cwhite, cblack;
    char *icon_name = "Xlife";
    XColor exact;

    if (!(disp = XOpenDisplay(getenv("DISPLAY"))))
	fatal("Can't open Display\n");

    if (XDefaultDepth(disp, screen) < STATEBITS)
	fatal("Not enough colors for compiled STATEBITS value\n");

    if (strstr(argv[0], "lifeconv") == argv[0] + strlen(argv[0]) - 8) {
      saveformat = 0;
      convmode++;
      if (argc < 3) {
le:
         fprintf(stderr, "Usage: lifeconv [-ACDIMPRSph] [-l pattern-lib] [-g n] filename\n");
         return 1;
      }
      for (i = 1; i < argc; i++)
        if (*argv[i] == '-')
              if (strchr("ACDIMPRSpvlg", argv[i][1]))
                  switch (argv[i][1]) {
                     case 'h':
                        goto le;
                     case 'g':
                        if (argc <= i + 1 || sscanf(argv[++i], "%d", &qgens) != 1) goto le;
                        break;
                     case 'l':
                        if (argc <= i + 1) goto le;
                        strcpy(matchlibfile, argv[++i]);
                        break;
                     case 'p':
                        oldp++;
                        argv[i][1] = 'P';
                     default:
                        if (saveformat != 0) goto le;
                        saveformat = argv[i][1];
                  }
              else
                 goto le;
        else
            initpat = argv[i];
    }
    else
      for (i = 1; i < argc; i++)
	if (strcmp(argv[i], "-geometry") == 0)
	    geomstring = argv[++i];
	else if (*argv[i] == '=')
	    geomstring = argv[i];
	else if (*argv[i] != '-') {
	    initpat = argv[i];
	    break;
	}
    if (convmode && initpat == 0) goto le;
    fileinit();
    initcells(&active);
    numcomments = 0;
    xpos = xorigin = lastx = STARTX;
    ypos = yorigin = lasty = STARTY;
    dispcoord = TRUE;
    multicolor = dispchanges = dispboxes = FALSE;
    dispmesh = 1;
    born = 8;
    live = 12;
    inputlength = width/FONTWIDTH;
    state = STOP;
    fname[0]='\0';
    strcpy(active_rules, "life");
    gentab();
    srandom(time((time_t *)NULL));

    screen = DefaultScreen(disp);
    rootw = RootWindow(disp, screen);
    cwhite = WhitePixel(disp, screen);
    cblack = BlackPixel(disp, screen);
    fcolor = cwhite;
    bcolor = cblack;

    hints.x = 0;
    hints.y = 0;
    width = DisplayWidth(disp,screen);
    height = DisplayHeight(disp,screen);

    hints.width = width * GRAB_FRACTION;
    hints.height = height * GRAB_FRACTION;

    hints.flags = PPosition | PSize;

    if (geomstring) {
	int result;

	result = XParseGeometry(geomstring,&hints.x,&hints.y,
				&hints.width,&hints.height);
	if (result & XNegative)
	    hints.x += (DisplayWidth(disp,screen) - hints.width) * GRAB_FRACTION;

	if (result & YNegative)
	    hints.y += (DisplayHeight(disp,screen) - hints.height) * GRAB_FRACTION;
	if (result & XValue || result & YValue) {
	    hints.flags |= USPosition;
	    hints.flags &= ~PPosition;
	}
	if (result & WidthValue || result & HeightValue) {
	    hints.flags |= USSize;
	    hints.flags &= ~PSize;
	}
    }

    mainw = XCreateSimpleWindow(disp, rootw,
		0, 0, hints.width, hints.height, 0, fcolor, bcolor);

    if (!mainw)
	fatal("Can't open main window");

    icon = XCreateBitmapFromData(disp, mainw, icon_bits, icon_width, icon_height);

    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
        fprintf (stderr, "structure allocation for iconName failed.\n");
        exit (6);
    }

    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.icon_pixmap = icon;
    wm_hints.flags = IconPixmapHint | StateHint | InputHint;

    class_hints.res_name =  argv[0];
    class_hints.res_class =  "Basicwin";

    XSetWMProperties(disp, mainw, 0, &iconName, argv, argc, &hints, &wm_hints, &class_hints );

    if (sizeof(fixedcolors)/sizeof(char *) < MAXCOLORS)
        fatal("Not enough colors for selected maximum number of states\n");
    for (i = 0; i < MAXCOLORS; i++) {
         if (!XAllocNamedColor(disp, DefaultColormap(disp, screen),
                             fixedcolors[i], &cellcolor[i], &exact))
             fatal("Color allocation failed!\n");
         xgcv.foreground = exact.pixel;
         if ((cellgc[i] = XCreateGC(disp, mainw,
                             GCForeground | GCBackground, &xgcv)) == 0)
             fatal("Color creation failed!\n");
    }
    black = cellcolor[0];
    white = cellcolor[1];

    /* text display is forced to black on white */
    xgcv.background = cwhite;
    xgcv.foreground = cblack;
    ntextgc = XCreateGC(disp, mainw, GCForeground | GCBackground, &xgcv);
    btextgc = XCreateGC(disp, mainw, GCForeground | GCBackground, &xgcv);
    xgcv.background = bcolor;
    xgcv.foreground = fcolor;

    /* create XOR GC for pivot display */
    tmp = xgcv.function;
    xgcv.function = GXinvert;
    xorgc = XCreateGC(disp, mainw, GCForeground |
                                   GCBackground | GCFunction, &xgcv);
    xgcv.function = tmp;

    if (!((nfont = XLoadQueryFont(disp, NORMALFONT)) && (bfont = XLoadQueryFont(disp, BOLDFONT))))
	fatal("Can't load font\n");
    XSetFont(disp, ntextgc, nfont->fid);
    XSetFont(disp, btextgc, bfont->fid);
    xgcv.function = GXcopy;
    xgcv.plane_mask = 1;

    cursor_data = XCreateBitmapFromData(disp, mainw, cursor_data_bits,
                                        cursor_data_width, cursor_data_height);
    cursor_mask = XCreateBitmapFromData(disp, mainw, cursor_mask_bits,
                                        cursor_mask_width, cursor_mask_height);
    cursor = XCreatePixmapCursor(disp, cursor_data, cursor_mask, &white,
                                        &black, cursor_data_x_hot, cursor_data_y_hot);
    XDefineCursor(disp, mainw, cursor);

    inputgc = XCreateGC(disp, mainw, GCFunction | GCPlaneMask, &xgcv);

    width = hints.width;
    height = hints.height;

    lifew = XCreateSimpleWindow(disp, mainw,
		0, 0, width - BORDERWIDTH*2, height - INPUTH - BORDERWIDTH*3,
                BORDERWIDTH,
		fcolor, bcolor);
    helpw = XCreateSimpleWindow(disp, mainw,
		0, 0, width - BORDERWIDTH*2, height - BORDERWIDTH*2,
                BORDERWIDTH,
		cblack, cwhite);
    rulew = XCreateSimpleWindow(disp, mainw,
		width - COORDW - RULEW - BORDERWIDTH*2,
		height - INPUTH - BORDERWIDTH*2,
                RULEW, INPUTH, BORDERWIDTH,
		cblack, cwhite);
    coordw = XCreateSimpleWindow(disp, mainw,
		width - COORDW - BORDERWIDTH*2,
		height - INPUTH - BORDERWIDTH*2,
                COORDW, INPUTH, BORDERWIDTH,
		cblack, cwhite);
    inputw = XCreateSimpleWindow(disp, mainw, 0,
                height - INPUTH - BORDERWIDTH*2, inputwwidth =
                width - COORDW - RULEW - BORDERWIDTH*3,
                INPUTH, BORDERWIDTH, cblack, cwhite);

    winat.win_gravity = SouthGravity;
    XChangeWindowAttributes(disp, inputw, CWWinGravity, &winat);

    XSelectInput(disp, mainw, KeyPressMask | ExposureMask | StructureNotifyMask);
    XSelectInput(disp, inputw, KeyPressMask | ButtonPressMask | ExposureMask);
    XSelectInput(disp, lifew, KeyPressMask | ButtonPressMask | Button1MotionMask | PointerMotionMask | Button3MotionMask | ButtonReleaseMask | ExposureMask);
    XSelectInput(disp, helpw, KeyPressMask | ExposureMask);
    XSelectInput(disp, rulew, KeyPressMask | ExposureMask | ButtonPressMask);
    XSelectInput(disp, coordw, KeyPressMask | ExposureMask);

    alloc_states(2);
    setscale(scale = 3);
    xpos -= shr(width >> 1, scale);
    ypos -= shr(height >> 1, scale);
    settimeout(delay = DELAY_FAST);
    /*
     * Only accept one pattern since it is highly unlikely that overlaying
     * n patterns is what you want to do
     */
    if (saveformat == 'C') {
        collect(initpat);
        return 0;
    }
    if (initpat) {
        loadx = xpos; 
        loady = ypos;
        txx = 1; 
        txy = 0; 
        tyx = 0; 
        tyy = 1; 
        add_loadreq(&loadqueue, 0, initpat, 0,
                    STARTX, STARTY, 1, 0, 0, 1);
        do_loadreq(loadqueue, &tentative);
    }

    if (convmode) {
       confirmload();
       while (qgens-- > 0)
          generate(&active);
       saveall(stdout, saveformat);
       return 0;
    }

    XMapWindow(disp, inputw);
    XMapWindow(disp, helpw);
    XMapWindow(disp, lifew);
    XMapWindow(disp, mainw);
    XMapWindow(disp, rulew);
    XMapWindow(disp, coordw);
    XLowerWindow(disp, helpw);

    showcoord(TRUE);
    showrules();

    for (;;) {
	while (XCheckMaskEvent(disp, KeyPressMask|ButtonPressMask|Button1MotionMask
	        |PointerMotionMask|Button3MotionMask|ButtonReleaseMask|ExposureMask
	        |StructureNotifyMask, &event)) {
	    switch(event.type) {
	      case KeyPress:
		XLookupString(&event.xkey, keybuf, 16, &ks, &kstat);
		if (!DoKeySymIn(ks))
                    if (event.xkey.window == lifew || !strchr("(){}[].JLOlx", *keybuf))
		        DoKeyIn(keybuf);
		break;
	      case MotionNotify:
                /* don't allow drawing until load is confirmed */
		if (!tentative.tiles)
		    Motion();
		break;
	      case ButtonPress:
		Button();
		break;
	      case ButtonRelease:
		Release();
		break;
	      case ConfigureNotify:
		DoResize();
		break;
	      case Expose:
		DoExpose(ClassifyWin(event.xexpose.window));
	    }
        }

	showcoord(FALSE);
	if (state == RUN || state == HIDE) {
            generate(&active);
            if (hideperiod > 1) {
                for (i = 1; i < hideperiod; i++)
	           generate(&active);
                if (state == RUN)
                   DoExpose(LIFEWIN);
            }
            else
	        redisplay();
	}
	else {
	    inputdelay.tv_sec = 0;
	    inputdelay.tv_usec = 100000;
	    select(32, 0, 0, 0, &inputdelay);
	}

	if (state == RUN) {
	    timeouttemp.tv_sec = timeout.tv_sec;
	    timeouttemp.tv_usec = timeout.tv_usec;
	    select(32, 0, 0, 0, &timeouttemp);
	}
    }
}

/* main.c ends here */
